import os
from time import sleep, time
from unittest import TestCase
from uuid import uuid4

from eventsourcing.application.multiprocess import MultiprocessRunner
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.application.system import MultiThreadedRunner, System
from eventsourcing.domain.model.events import assert_event_handlers_empty, clear_event_handlers
from eventsourcing.tests.test_process import ExampleAggregate
from eventsourcing.tests.test_system_fixtures import Examples, Order, Orders, Payment, Payments, Reservation, \
    Reservations, create_new_order, set_db_uri


class TestSystem(TestCase):
    infrastructure_class = SQLAlchemyApplication

    def test___getattr__(self):
        system = System(Orders | Reservations | Orders,
                        Orders | Payments | Orders,
                        setup_tables=True,
                        infrastructure_class=self.infrastructure_class
                        )
        with system.orders as app:
            self.assertIsInstance(app, Orders)
            self.assertEqual(app, system.orders)

        with system.payments as app:
            self.assertIsInstance(app, Payments)
            self.assertEqual(app, system.payments)

        with system.reservations as app:
            self.assertIsInstance(app, Reservations)
            self.assertEqual(app, system.reservations)

        with self.assertRaises(AttributeError):
            with system.notaprocess as _: pass

    def test_singlethreaded_runner_with_multiapp_system(self):
        system = System(Orders | Reservations | Orders,
                        Orders | Payments | Orders,
                        setup_tables=True,
                        infrastructure_class=self.infrastructure_class
                        )

        with system as runner:
            # Create new Order aggregate.
            order_id = create_new_order()

            # Check the order is reserved and paid.
            repository = runner.orders.repository
            assert repository[order_id].is_reserved
            assert repository[order_id].is_paid

    def test_singlethreaded_runner_with_direct_query(self):
        system = System(Orders | Reservations | Orders,
                        Orders | Payments | Orders,
                        setup_tables=True,
                        infrastructure_class=self.infrastructure_class,
                        use_direct_query_if_available=True,
                        )

        with system:
            # Create new Order aggregate.
            order_id = create_new_order()

            # Check the order is reserved and paid.
            repository = system.orders.repository
            assert repository[order_id].is_reserved
            assert repository[order_id].is_paid

    def test_multithreaded_runner_with_singleapp_system(self):

        system = System(Examples | Examples,
                        setup_tables=True,
                        infrastructure_class=self.infrastructure_class)

        self.set_db_uri()

        with MultiThreadedRunner(system):

            app = system.examples

            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            assert aggregate.id in app.repository

            # Check the aggregate is moved on.
            retries = 50
            while not app.repository[aggregate.id].is_moved_on:

                sleep(0.1)
                retries -= 1
                assert retries, "Failed to move"

    def test_multithreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class
        )

        self.set_db_uri()

        with MultiThreadedRunner(system):

            started = time()

            orders = system.orders

            # Create new orders.
            num_orders = 10
            order_ids = []
            for i in range(num_orders):
                order_id = create_new_order()
                order_ids.append(order_id)

            retries = num_orders
            for order_id in order_ids:
                # while not orders.repository[order_id].is_reserved:
                #     sleep(0.1)
                #     retries -= 1
                #     assert retries, "Failed set order.is_reserved"

                while retries and not orders.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid"

            print(f"Duration: { time() - started :.4f}s")

    def test_clocked_multithreaded_runner_with_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class
        )

        self.set_db_uri()

        clock_speed = 10
        tick_interval = 1 / clock_speed
        with MultiThreadedRunner(system, clock_speed=clock_speed):

            started = time()

            orders = system.orders

            # Create a new order.
            num_orders = 10
            order_ids = []
            for i in range(num_orders):
                order_id = create_new_order()
                order_ids.append(order_id)
                # sleep(tick_interval / 3)
                # sleep(tick_interval * 10)

            retries = 30 * num_orders
            num_completed = 0
            for order_id in order_ids:
                while retries and not orders.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid (after %s completed)" % num_completed
                num_completed += 1

        print(f"Duration: { time() - started :.4f}s")

    def test_multiprocessing_singleapp_system(self):

        system = System(Examples | Examples,
                        setup_tables=True,
                        infrastructure_class=self.infrastructure_class)

        self.set_db_uri()

        self.close_connections_before_forking()

        with MultiprocessRunner(system), system.examples as app:

            aggregate = ExampleAggregate.__create__()
            aggregate.__save__()

            assert aggregate.id in app.repository

            # Check the aggregate is moved on.
            retries = 50
            while not app.repository[aggregate.id].is_moved_on:

                sleep(0.1)
                retries -= 1
                assert retries, "Failed to move"

    def test_multiprocessing_multiapp_system(self):
        system = System(
            Orders | Reservations | Orders,
            Orders | Payments | Orders,
            setup_tables=True,
            infrastructure_class=self.infrastructure_class
        )

        self.set_db_uri()

        with system.construct_app(Orders) as app:
            # Create a new order.
            order_id = create_new_order()
            # Check new order exists in the repository.
            assert order_id in app.repository

        self.close_connections_before_forking()

        with MultiprocessRunner(system):

            with system.construct_app(Orders) as app:
                retries = 50
                while not app.repository[order_id].is_reserved:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved"

                while retries and not app.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid"

    def test_multipipeline_multiprocessing_multiapp(self):

        self.set_db_uri()

        system = System(
            (Orders, Reservations, Orders, Payments, Orders),
            setup_tables=True,
            infrastructure_class=self.infrastructure_class
        )

        num_pipelines = 2

        pipeline_ids = range(num_pipelines)

        multiprocess = MultiprocessRunner(system, pipeline_ids=pipeline_ids)

        num_orders_per_pipeline = 5
        order_ids = []

        self.close_connections_before_forking()

        # Start multiprocessing system.
        with multiprocess, system.construct_app(Orders) as orders:

            # Create some new orders.
            for _ in range(num_orders_per_pipeline):

                for pipeline_id in pipeline_ids:

                    orders.change_pipeline(pipeline_id)

                    order_id = create_new_order()
                    order_ids.append(order_id)

                    sleep(0.05)

            # Wait for orders to be reserved and paid.
            retries = 10 + 10 * num_orders_per_pipeline * len(pipeline_ids)
            for i, order_id in enumerate(order_ids):

                while not orders.repository[order_id].is_reserved:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_reserved {} ({})".format(order_id, i)

                while retries and not orders.repository[order_id].is_paid:
                    sleep(0.1)
                    retries -= 1
                    assert retries, "Failed set order.is_paid ({})".format(i)

            # Calculate timings from event timestamps.
            order_aggregates = [orders.repository[oid] for oid in order_ids]
            first_timestamp = min([o.__created_on__ for o in order_aggregates])
            last_timestamp = max([o.__last_modified__ for o in order_aggregates])
            duration = last_timestamp - first_timestamp
            rate = len(order_ids) / float(duration)
            period = 1 / rate
            print("Orders system processed {} orders in {:.3f}s at rate of {:.1f} "
                  "orders/s, {:.3f}s each".format(len(order_ids), duration, rate, period))

            # Print min, average, max duration.
            durations = [o.__last_modified__ - o.__created_on__ for o in order_aggregates]
            print("Min order processing time: {:.3f}s".format(min(durations)))
            print("Mean order processing time: {:.3f}s".format(sum(durations) / len(durations)))
            print("Max order processing time: {:.3f}s".format(max(durations)))

    def set_db_uri(self):
        set_db_uri()

    def close_connections_before_forking(self):
        # Used for closing Django connection before multiprocessing module forks the OS process.
        pass

    def test_payments_policy(self):
        # Prepare fake repository with a real Order aggregate.
        order = Order.__create__()
        fake_repository = {order.id: order}

        # Check policy makes payment whenever order is reserved.
        event = Order.Reserved(originator_id=order.id, originator_version=1)

        payment = Payments.policy(repository=fake_repository, event=event)
        assert isinstance(payment, Payment), payment
        assert payment.order_id == order.id

    def test_orders_policy(self):
        # Prepare fake repository with a real Order aggregate.
        order = Order.__create__()
        fake_repository = {order.id: order}

        # Check order is not reserved.
        assert not order.is_reserved

        # Reservation created.
        event = Reservation.Created(originator_id=uuid4(), originator_topic='', order_id=order.id)
        Orders.policy(repository=fake_repository, event=event)

        # Check order is reserved.
        assert order.is_reserved

    def tearDown(self):
        assert_event_handlers_empty()
        clear_event_handlers()
        try:
            del (os.environ['DB_URI'])
        except KeyError:
            pass



from eventsourcing.infrastructure.integersequencegenerators.base import SimpleIntegerSequenceGenerator
from eventsourcing.infrastructure.integersequencegenerators.redisincr import RedisIncr
from eventsourcing.tests.base import AbstractTestCase


class IntegerSequenceGeneratorTestCase(AbstractTestCase):
    generator_class = None

    def test_iterable(self):
        g = self.generator_class()
        limit = 500
        for i, j in enumerate(g):
            self.assertEqual(i, j)
            if i == limit:
                break
        else:
            self.fail('There were no items in the sequence')
        self.assertEqual(i, limit)

    def test_iterator(self):
        g = self.generator_class()
        limit = 500
        for i in range(limit):
            j = next(g)
            self.assertEqual(i, j)


class TestSimpleIntegerSequenceGenerator(IntegerSequenceGeneratorTestCase):
    generator_class = SimpleIntegerSequenceGenerator


class TestRedisIncr(IntegerSequenceGeneratorTestCase):
    generator_class = RedisIncr



from functools import singledispatch, wraps
from inspect import isfunction


def applicationpolicy(arg=None):
    """
    Decorator for application policy method.

    Allows policy to be built up from methods
    registered for different event classes.
    """

    def _mutator(func):
        wrapped = singledispatch(func)

        @wraps(wrapped)
        def wrapper(*args, **kwargs):
            event = kwargs.get('event') or args[-1]
            return wrapped.dispatch(type(event))(*args, **kwargs)

        wrapper.register = wrapped.register

        return wrapper

    assert isfunction(arg), arg
    return _mutator(arg)



import os

from eventsourcing.application.notificationlog import RecordManagerNotificationLog
from eventsourcing.application.pipeline import Pipeable
from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.infrastructure.base import DEFAULT_PIPELINE_ID
from eventsourcing.infrastructure.eventsourcedrepository import EventSourcedRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.factory import InfrastructureFactory
from eventsourcing.infrastructure.sequenceditem import StoredEvent
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.utils.cipher.aes import AESCipher
from eventsourcing.utils.random import decode_bytes


class SimpleApplication(Pipeable):
    """
    Base class for event sourced applications.

    Constructs infrastructure objects such as the repository and
    event store, and also the notification log which presents the
    application state as a sequence of events.

    Needs actual infrastructure classes.
    """
    infrastructure_factory_class = InfrastructureFactory
    is_constructed_with_session = False

    record_manager_class = None
    stored_event_record_class = None
    snapshot_record_class = None

    sequenced_item_class = None
    sequenced_item_mapper_class = None
    json_encoder_class = None
    json_decoder_class = None

    persist_event_type = None
    notification_log_section_size = None
    use_cache = False

    event_store_class = EventStore
    repository_class = EventSourcedRepository

    def __init__(self, name='', persistence_policy=None, persist_event_type=None,
                 cipher_key=None, sequenced_item_class=None, sequenced_item_mapper_class=None,
                 record_manager_class=None, stored_event_record_class=None, event_store_class=None,
                 snapshot_record_class=None, setup_table=True,
                 contiguous_record_ids=True, pipeline_id=DEFAULT_PIPELINE_ID, json_encoder_class=None,
                 json_decoder_class=None, notification_log_section_size=None):
        self._datastore = None
        self._event_store = None
        self._repository = None
        self.infrastructure_factory = None

        self.name = name or type(self).__name__.lower()

        self.notification_log_section_size = notification_log_section_size

        self.sequenced_item_class = sequenced_item_class \
                                    or type(self).sequenced_item_class \
                                    or StoredEvent

        self.sequenced_item_mapper_class = sequenced_item_mapper_class \
                                           or type(self).sequenced_item_mapper_class \
                                           or SequencedItemMapper

        self.record_manager_class = record_manager_class or type(self).record_manager_class

        self.event_store_class = event_store_class or type(self).event_store_class

        self._stored_event_record_class = stored_event_record_class

        self._snapshot_record_class = snapshot_record_class

        self.json_encoder_class = json_encoder_class or type(self).json_encoder_class
        self.json_decoder_class = json_decoder_class or type(self).json_decoder_class
        self.persist_event_type = persist_event_type or type(self).persist_event_type

        self.contiguous_record_ids = contiguous_record_ids
        self.pipeline_id = pipeline_id
        self.persistence_policy = persistence_policy

        self.construct_cipher(cipher_key)

        if self.record_manager_class or self.infrastructure_factory_class.record_manager_class:
            self.construct_infrastructure()

            if setup_table:
                self.setup_table()
            self.construct_notification_log()

            if self.persistence_policy is None:
                self.construct_persistence_policy()

    @property
    def datastore(self):
        return self._datastore

    @property
    def event_store(self) -> EventStore:
        return self._event_store

    @property
    def repository(self) -> EventSourcedRepository:
        return self._repository

    def construct_cipher(self, cipher_key):
        cipher_key = decode_bytes(cipher_key or os.getenv('CIPHER_KEY', ''))
        self.cipher = AESCipher(cipher_key) if cipher_key else None

    def construct_infrastructure(self, *args, **kwargs):
        self.infrastructure_factory = self.construct_infrastructure_factory(*args, **kwargs)
        self.construct_datastore()
        self.construct_event_store()
        self.construct_repository()

    def construct_infrastructure_factory(self, *args, **kwargs):
        """
        :rtype: InfrastructureFactory
        """
        factory_class = self.infrastructure_factory_class
        assert issubclass(factory_class, InfrastructureFactory)

        integer_sequenced_record_class = self._stored_event_record_class or self.stored_event_record_class
        snapshot_record_class = self._snapshot_record_class or self.snapshot_record_class

        return factory_class(
            record_manager_class=self.record_manager_class,
            integer_sequenced_record_class=integer_sequenced_record_class,
            snapshot_record_class=snapshot_record_class,
            sequenced_item_class=self.sequenced_item_class,
            sequenced_item_mapper_class=self.sequenced_item_mapper_class,
            contiguous_record_ids=self.contiguous_record_ids,
            application_name=self.name,
            pipeline_id=self.pipeline_id,
            event_store_class=self.event_store_class,
            *args, **kwargs
        )

    def construct_datastore(self):
        self._datastore = self.infrastructure_factory.construct_datastore()

    def construct_event_store(self):
        self._event_store = self.infrastructure_factory.construct_integer_sequenced_event_store(self.cipher)

    def construct_repository(self, **kwargs):
        self._repository = self.repository_class(
            event_store=self.event_store,
            use_cache=self.use_cache,
            **kwargs
        )

    def setup_table(self):
        # Setup the database table using event store's record class.
        if self.datastore is not None:
            self.datastore.setup_table(
                self.event_store.record_manager.record_class
            )

    def drop_table(self):
        # Drop the database table using event store's record class.
        if self.datastore is not None:
            self.datastore.drop_table(
                self.event_store.record_manager.record_class
            )

    def construct_notification_log(self):
        self.notification_log = RecordManagerNotificationLog(
            self.event_store.record_manager,
            section_size=self.notification_log_section_size
        )

    def construct_persistence_policy(self):
        self.persistence_policy = PersistencePolicy(
            event_store=self.event_store,
            persist_event_type=self.persist_event_type
        )

    def change_pipeline(self, pipeline_id):
        self.pipeline_id = pipeline_id
        self.event_store.record_manager.pipeline_id = pipeline_id

    def close(self):
        # Close the persistence policy.
        if self.persistence_policy is not None:
            self.persistence_policy.close()

        # Close database connection.
        if self.datastore is not None:
            self.datastore.close_connection()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def reset_connection_after_forking(cls):
        pass

    @classmethod
    def mixin(cls, infrastructure_class):
        return type(cls.__name__, (infrastructure_class, cls), {})


class ApplicationWithConcreteInfrastructure(SimpleApplication):
    """
    Subclasses have actual infrastructure.
    """



from base64 import b64encode, b64decode

import os


def encode_random_bytes(num_bytes):
    """
    Generates random bytes, encoded as Base64 unicode string.

    :param int num_bytes: Number of random bytes to generate.
    :return: Random bytes of specified length, encoded as Base64 unicode string.
    :type: str
    """
    return encode_bytes(random_bytes(num_bytes))


def random_bytes(num_bytes):
    """
    Generates random bytes.

    :param int num_bytes: Number of random bytes to generate.
    :return: Random bytes of specified length.
    :type: bytes
    """
    return os.urandom(num_bytes)


def encode_bytes(bytes):
    """
    Encodes bytes as Base64 unicode string.

    """
    return b64encode(bytes).decode('utf-8')


def decode_bytes(s):
    """
    Decodes bytes from Base64 encoded unicode string.

    :param str s: Base64 string.
    :returns: Bytes that were encoded with Base64.
    :rtype: bytes
    """
    return b64decode(s.encode('utf-8'))



import os

os.environ['DJANGO_SETTINGS_MODULE'] = 'eventsourcing.tests.djangoproject.djangoproject.settings'

import django

django.setup()

import unittest
from time import sleep

from django.core.management import call_command
from django.test import TransactionTestCase

from eventsourcing.infrastructure.django.utils import close_django_connection
from eventsourcing.infrastructure.django.apps import DjangoConfig
from eventsourcing.infrastructure.django.factory import DjangoInfrastructureFactory
from eventsourcing.tests.sequenced_item_tests import base


class DjangoTestCase(TransactionTestCase):
    infrastructure_factory_class = DjangoInfrastructureFactory
    contiguous_record_ids = True

    def setUp(self):
        super(DjangoTestCase, self).setUp()
        # Setup tables (there isn't a Django datastore object, but we can do it like this).
        call_command('migrate', verbosity=0, interactive=False)
        sleep(1)

    def tearDown(self):
        # Drop tables (there isn't a Django datastore object, but we can do it like this).
        # call_command('migrate', 'django', 'zero', verbosity=0, interactive=False)
        super(DjangoTestCase, self).tearDown()

    def construct_datastore(self):
        pass

    def close_connections_before_forking(self):
        # If connection is already made close it.
        close_django_connection()


class TestDjangoRecordManagerWithIntegerSequences(DjangoTestCase, base.IntegerSequencedRecordTestCase):
    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestDjangoRecordManagerWithoutContiguousRecordIDs(DjangoTestCase, base.IntegerSequencedRecordTestCase):
    contiguous_record_ids = False

    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestDjangoRecordManagerWithTimestampSequences(DjangoTestCase, base.TimestampSequencedItemTestCase):
    def construct_record_manager(self):
        return self.construct_timestamp_sequenced_record_manager()


# class WithDjangoRecordManagers(DjangoTestCase, WithActiveRecordManagers):
#     def construct_entity_record_manager(self):
#         return construct_integer_sequenced_record_manager()
#
#     def construct_log_record_manager(self):
#         return construct_timestamp_sequenced_record_manager()
#
#     def construct_snapshot_record_manager(self):
#         return construct_snapshot_record_manager()
#

# class TestSimpleSequencedItemIteratorWithDjango(WithDjangoRecordManagers,
#                                                 SimpleSequencedItemteratorTestCase):
#     pass
#
#
# class TestThreadedSequencedItemIteratorWithDjango(WithDjangoRecordManagers,
#                                                   ThreadedSequencedItemIteratorTestCase):
#     pass

class TestConfigClass(unittest.TestCase):

    def test(self):
        self.assertEqual('eventsourcing.infrastructure.django', DjangoConfig.name)



from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper


class SequencedItemMapperForPopo(SequencedItemMapper):
    def get_event_class_and_attrs(self, topic, data):
        return topic, data

    def get_item_topic_and_state(self, domain_event_class, event_attrs):
        return domain_event_class, event_attrs



from unittest import TestCase

from eventsourcing.tests.sequenced_item_tests.test_django_record_manager import DjangoTestCase

from eventsourcing.application.django import DjangoApplication
from eventsourcing.application.notificationlog import NotificationLogReader
from eventsourcing.application.snapshotting import SnapshottingApplication
from eventsourcing.application.sqlalchemy import SQLAlchemyApplication
from eventsourcing.domain.model.events import assert_event_handlers_empty, DomainEvent
from eventsourcing.tests.core_tests.test_aggregate_root import ExampleAggregateRoot
from eventsourcing.utils.random import encode_random_bytes


class TestSimpleApplication(TestCase):

    application_class = SQLAlchemyApplication

    def test(self):
        with self.get_application() as app:

            # Start with a new table.
            app.drop_table()
            app.drop_table()
            app.setup_table()
            app.setup_table()

            # Check the application's persistence policy,
            # repository, and event store, are working.
            aggregate = ExampleAggregateRoot.__create__()
            aggregate.__save__()
            self.assertTrue(aggregate.id in app.repository)

            # Check the notifications.
            reader = NotificationLogReader(app.notification_log)
            notifications = reader.read_list()
            self.assertEqual(1, len(notifications))
            topic = 'eventsourcing.tests.core_tests.test_aggregate_root#ExampleAggregateRoot.Created'
            self.assertEqual(topic, notifications[0]['topic'])

            app.drop_table()

    def get_application(self):
        return self.application_class(
            cipher_key=encode_random_bytes(16),
            persist_event_type=DomainEvent,

        )

    def tearDown(self):
        # Check the close() method leaves everything unsubscribed.
        assert_event_handlers_empty()


class TestDjangoApplication(DjangoTestCase, TestSimpleApplication):
    application_class = DjangoApplication


class TestSnapshottingApplication(TestSimpleApplication):
    application_class = SnapshottingApplication.mixin(SQLAlchemyApplication)



import os
from datetime import datetime, timedelta
from decimal import Decimal
from time import sleep
from unittest.case import TestCase
from uuid import uuid1

import sys

from eventsourcing.utils.random import encode_random_bytes, decode_bytes
from eventsourcing.utils.times import decimaltimestamp_from_uuid, utc_timezone, decimaltimestamp


class TestUtils(TestCase):
    def test_decimaltimestamp(self):
        # Check a given float is correctly converted to a Decimal.
        self.assertEqual(decimaltimestamp(123456789.1234560), Decimal('123456789.123456'))
        self.assertEqual(decimaltimestamp(1234567890.1234560), Decimal('1234567890.123456'))

        # Generate a series of timestamps.
        timestamps = [decimaltimestamp() for _ in range(100)]

        # Check series doesn't decrease at any point.
        last = timestamps[0]
        for timestamp in timestamps[1:]:
            self.assertLessEqual(last, timestamp)
            last = timestamp

    def test_decimaltimestamp_from_uuid(self):
        # Check can convert a UUID to a Decimal.
        self.assertIsInstance(decimaltimestamp_from_uuid(uuid1()), Decimal)
        # Check can convert a UUID hex to a Decimal.
        self.assertIsInstance(decimaltimestamp_from_uuid(uuid1().hex), Decimal)

        # Generate a series of timestamps.
        timestamps = [decimaltimestamp_from_uuid(uuid1()) for _ in range(100)]

        # Check series doesn't decrease at any point.
        last = timestamps[0]
        for timestamp in timestamps[1:]:
            self.assertLessEqual(last, timestamp)
            last = timestamp

        # Check last timestamp is greater than the first.
        self.assertGreater(timestamps[-1], timestamps[0])

    def test_decimaltimestamp_corresponds_with_decimaltimestamp_from_uuid(self):
        if os.getenv('TRAVIS_PYTHON_VERSION') in ['3.6', '3.7', '3.7-dev', 'pypy3.5']:
            self.skipTest("Somehow this fails on Travis dist:xenial.")

            # This is the weird error that happens in Python 3.7.1 on Travis Xenial dist.
            # Why does the second timestamp "happen" more than one second before the first?
            # Perhaps UUID1 doesn't actually use time.time() sometimes? Maybe it was a bug in v3.7.1?
            """
FAIL: test_decimaltimestamp_corresponds_with_decimaltimestamp_from_uuid (
eventsourcing.tests.core_tests.test_utils.TestUtils)
----------------------------------------------------------------------
Traceback (most recent call last):
  File "/home/travis/build/johnbywater/eventsourcing/eventsourcing/tests/core_tests/test_utils.py", line 61,
  in test_decimaltimestamp_corresponds_with_decimaltimestamp_from_uuid
    self.assertLess(timestamp1, timestamp2)
AssertionError: Decimal('1561464862.322443') not less than Decimal('1561464861.100940')
            """

        timestamp1 = decimaltimestamp()
        sleep(0.000001)
        uuid_1 = uuid1()
        sleep(0.000001)
        timestamp3 = decimaltimestamp()
        sleep(0.000001)
        timestamp2 = decimaltimestamp_from_uuid(uuid_1)
        self.assertLess(timestamp1, timestamp2)
        self.assertLess(timestamp2, timestamp3)

    def test_utc(self):
        now = datetime.now(tz=utc_timezone)
        self.assertEqual(utc_timezone.utcoffset(now), timedelta(0))
        expected_dst = None if int(sys.version[0]) > 2 else timedelta(0)
        self.assertEqual(utc_timezone.dst(now), expected_dst)

    def test_encode_random_bytes(self):
        key = encode_random_bytes(num_bytes=16)
        self.assertEqual(len(decode_bytes(key)), 16)

        key = encode_random_bytes(num_bytes=24)
        self.assertEqual(len(decode_bytes(key)), 24)

        key = encode_random_bytes(num_bytes=32)
        self.assertEqual(len(decode_bytes(key)), 32)



from abc import ABC, abstractmethod

from eventsourcing.domain.model.events import EventWithOriginatorID, EventWithOriginatorVersion, EventWithTimestamp


class AbstractSnapshop(ABC):
    @property
    @abstractmethod
    def topic(self):
        """
        Path to the class of the snapshotted entity.
        """

    @property
    @abstractmethod
    def state(self):
        """
        State of the snapshotted entity.
        """

    @property
    @abstractmethod
    def originator_id(self):
        """
        ID of the snapshotted entity.
        """

    @property
    @abstractmethod
    def originator_version(self):
        """
        Version of the last event applied to the entity.
        """


class Snapshot(EventWithTimestamp, EventWithOriginatorVersion, EventWithOriginatorID, AbstractSnapshop):
    def __init__(self, originator_id, originator_version, topic, state):
        super(Snapshot, self).__init__(
            originator_id=originator_id,
            originator_version=originator_version,
            topic=topic,
            state=state,
        )

    @property
    def topic(self):
        """
        Path to the class of the snapshotted entity.
        """
        return self.__dict__['topic']

    @property
    def state(self):
        """
        State of the snapshotted entity.
        """
        return self.__dict__['state']



from eventsourcing.tests.datastore_tests import test_sqlalchemy
from eventsourcing.tests.sequenced_item_tests import base


class SQLAlchemyRecordManagerTestCase(test_sqlalchemy.SQLAlchemyDatastoreTestCase,
                                      base.WithRecordManagers):
    """
    Base class for test cases that need record manager with SQLAlchemy.
    """


class TestSQLAlchemyRecordManagerWithIntegerSequences(SQLAlchemyRecordManagerTestCase,
                                                      base.IntegerSequencedRecordTestCase):
    """
    Test case for integer sequenced record manager with SQLAlchemy.
    """


class TestSQLAlchemyRecordManagerWithTimestampSequences(SQLAlchemyRecordManagerTestCase,
                                                        base.TimestampSequencedItemTestCase):
    """
    Test case for timestamp sequenced record manager with SQLAlchemy.
    """


class TestSimpleIteratorWithSQLAlchemy(SQLAlchemyRecordManagerTestCase,
                                       base.SequencedItemIteratorTestCase):
    """
    Test case for simple iterator in record manager with SQLAlchemy.
    """


class TestThreadedIteratorWithSQLAlchemy(SQLAlchemyRecordManagerTestCase,
                                         base.ThreadedSequencedItemIteratorTestCase):
    """
    Test case for threaded iterator in record manager with SQLAlchemy.
    """
    use_named_temporary_file = True



import os
import sys
from os.path import dirname, join
from subprocess import PIPE, Popen
from tempfile import NamedTemporaryFile
from unittest.case import TestCase

import eventsourcing
from eventsourcing.domain.model.events import assert_event_handlers_empty, clear_event_handlers
from eventsourcing.infrastructure.sqlalchemy.datastore import SQLAlchemyDatastore, SQLAlchemySettings
from eventsourcing.infrastructure.sqlalchemy.records import Base

base_dir = dirname(dirname(os.path.abspath(eventsourcing.__file__)))


class TestDocs(TestCase):
    def tearDown(self):
        clear_event_handlers()

        # Need to drop the stored events, because the __main__#Order topics
        # mess up the multiprocessing tests (because the Orders application
        # has the same ID so events from one test cause another to fail).
        os.environ['DB_URI'] = 'mysql+pymysql://{}:{}@{}/eventsourcing?charset=utf8mb4&binary_prefix=true'.format(
            os.getenv('MYSQL_USER', 'root'),
            os.getenv('MYSQL_PASSWORD', ''),
            os.getenv('MYSQL_HOST', '127.0.0.1'),
        )
        database = SQLAlchemyDatastore(settings=SQLAlchemySettings())
        database.setup_connection()
        # Might as well drop everything....
        Base.metadata.drop_all(database._engine)

        del (os.environ['DB_URI'])

    def test_readme(self):
        self._out = ''
        path = join(base_dir, 'README.md')
        if not os.path.exists(path):
            self.skipTest("Skipped test, README file not found: {}".format(path))
        self.check_code_snippets_in_file(path)

    def test_docs(self):

        skipped = [
            # 'deployment.rst'
        ]

        self._out = ''
        docs_path = os.path.join(base_dir, 'docs')

        if not os.path.exists(docs_path):
            self.skipTest("Skipped test, docs folder not found: {}".format(docs_path))

        file_paths = []
        for dirpath, _, filenames in os.walk(docs_path):
            for name in filenames:
                if name in skipped:
                    continue
                if name.endswith('.rst'):
                # if name.endswith('domainmodel.rst'):
                # if name.endswith('quick_start.rst'):
                # if name.endswith('aggregates_in_ddd.rst'):
                # if name.endswith('example_application.rst'):
                # if name.endswith('everything.rst'):
                # if name.endswith('infrastructure.rst'):
                # if name.endswith('application.rst'):
                # if name.endswith('snapshotting.rst'):
                # if name.endswith('notifications.rst'):
                # if name.endswith('projections.rst'):
                # if name.endswith('deployment.rst'):
                # if name.endswith('process.rst'):
                    file_paths.append(os.path.join(docs_path, dirpath, name))

        file_paths = sorted(file_paths)
        failures = []
        passed = []
        failed = []
        print("Testing code snippets in docs:")
        for path in file_paths:
            print(path)
        print('')
        for path in file_paths:
            # print("Testing code snippets in file: {}".format(path))
            try:
                self.check_code_snippets_in_file(path)
                assert_event_handlers_empty()
            except self.failureException as e:
                failures.append(e)
                failed.append(path)
                print(str(e).strip('\n'))
                print('FAIL')
                print('')
            else:
                passed.append(path)
                print('PASS')
                print('')
            finally:
                os.environ['SALT_FOR_DATA_INTEGRITY'] = ''

        print("{} failed, {} passed".format(len(failed), len(passed)))

        if failures:
            raise failures[0]

    def check_code_snippets_in_file(self, doc_path):
        # Extract lines of Python code from the README.md file.

        lines = []
        num_code_lines = 0
        num_code_lines_in_block = 0
        is_code = False
        is_md = False
        is_rst = False
        last_line = ''
        with open(doc_path) as doc_file:
            for orig_line in doc_file:
                line = orig_line.strip('\n')
                if line.startswith('```python'):
                    # Start markdown code block.
                    if is_rst:
                        self.fail("Markdown code block found after restructured text block in same file.")
                    is_code = True
                    is_md = True
                    line = ''
                    num_code_lines_in_block = 0
                elif is_code and is_md and line.startswith('```'):
                    # Finish markdown code block.
                    if not num_code_lines_in_block:
                        self.fail("No lines of code in block")
                    is_code = False
                    line = ''
                elif is_code and is_rst and line.startswith('```'):
                    # Can't finish restructured text block with markdown.
                    self.fail("Restructured text block terminated with markdown format '```'")
                elif line.startswith('.. code:: python') and 'exclude-when-testing' not in last_line:
                    # Start restructured text code block.
                    if is_md:
                        self.fail("Restructured text code block found after markdown block in same file.")
                    is_code = True
                    is_rst = True
                    line = ''
                    num_code_lines_in_block = 0
                elif is_code and is_rst and line and not line.startswith(' '):
                    # Finish restructured text code block.
                    if not num_code_lines_in_block:
                        self.fail("No lines of code in block")
                    is_code = False
                    line = ''
                elif is_code:
                    # Process line in code block.
                    if is_rst:
                        # Restructured code block normally indented with four spaces.
                        if len(line.strip()):
                            if not line.startswith('    '):
                                self.fail("Code line needs 4-char indent: {}: {}".format(repr(line), doc_path))
                            # Strip four chars of indentation.
                            line = line[4:]

                    if len(line.strip()):
                        num_code_lines_in_block += 1
                        num_code_lines += 1
                else:
                    line = ''
                lines.append(line)
                last_line = orig_line

        print("{} lines of code in {}".format(num_code_lines, doc_path))

        # Write the code into a temp file.
        tempfile = NamedTemporaryFile('w+')
        temp_path = tempfile.name
        tempfile.writelines("\n".join(lines) + '\n')
        tempfile.flush()

        # Run the code and catch errors.
        p = Popen([sys.executable, temp_path], stdout=PIPE, stderr=PIPE)
        out, err = p.communicate()
        out = out.decode('utf8')
        err = err.decode('utf8')
        out = out.replace(temp_path, doc_path)
        err = err.replace(temp_path, doc_path)
        exit_status = p.wait()

        print(out)
        print(err)

        # Check for errors running the code.
        if exit_status:
            self.fail(out + err)

        # Close (deletes) the tempfile.
        tempfile.close()



# import itertools
# from collections import defaultdict
#
# from eventsourcing.exceptions import ConcurrencyError, DatasourceOperationError
# from eventsourcing.infrastructure.eventstore import AbstractStoredEventRepository
# from eventsourcing.infrastructure.transcoding import EntityVersion
# from eventsourcing.utils.times import timestamp_from_uuid
#
#
# class PythonObjectsStoredEventRepository(AbstractStoredEventRepository):
#     def __init__(self, *args, **kwargs):
#         super(PythonObjectsStoredEventRepository, self).__init__(*args, **kwargs)
#         self._by_id = {}
#         self._by_stored_entity_id = {}
#         self._originator_versions = defaultdict(lambda: defaultdict(lambda: itertools.chain([0], itertools.cycle([
# 1]))))
#         self._event_id_by_originator_version_id = {}
#
#     def write_version_and_event(self, new_stored_event, new_version_number=None, max_retries=3,
#                                 artificial_failure_rate=0):
#
#         if artificial_failure_rate:
#             raise DatasourceOperationError('Artificial failure')
#
#         # Put the event in the various dicts.
#         stored_entity_id = new_stored_event.entity_id
#         if self.always_write_originator_version and new_version_number is not None:
#             versions = self._originator_versions[stored_entity_id]
#             if next(versions[new_version_number]) != 0:
#                 raise ConcurrencyError("New version {} for entity {} already exists"
#                                        "".format(new_version_number, stored_entity_id))
#             originator_version_id = self.make_originator_version_id(stored_entity_id, new_version_number)
#             self._event_id_by_originator_version_id[originator_version_id] = new_stored_event.event_id
#
#         # Remove entity if it's a discarded event.
#         if new_stored_event.event_topic.endswith('Discarded'):
#             self.remove_entity(stored_entity_id)
#
#         # Otherwise add event to entity's list of events.
#         else:
#             # Index by entity ID.
#             if stored_entity_id not in self._by_stored_entity_id:
#                 self._by_stored_entity_id[stored_entity_id] = []
#             self._by_stored_entity_id[stored_entity_id].append(new_stored_event)
#
#             # Index by event ID.
#             self._by_id[new_stored_event.event_id] = new_stored_event
#
#     def get_originator_version(self, stored_entity_id, version_number):
#         versions = self._originator_versions[stored_entity_id]
#         if version_number not in versions:
#             self.raise_originator_version_not_found(stored_entity_id, version_number)
#         originator_version_id = self.make_originator_version_id(stored_entity_id, version_number)
#         return EntityVersion(
#             originator_version_id=originator_version_id,
#             event_id=self._event_id_by_originator_version_id[originator_version_id]
#         )
#
#     def remove_entity(self, stored_entity_id):
#         if stored_entity_id in self._by_stored_entity_id:
#             for stored_event in self._by_stored_entity_id.pop(stored_entity_id):
#                 del (self._by_id[stored_event.event_id])
#         if self.always_write_originator_version:
#             del (self._originator_versions[stored_entity_id])
#
#     def get_stored_events(self, stored_entity_id, after=None, until=None, limit=None, query_ascending=True,
#                           results_ascending=True, include_after_when_ascending=False,
#                           include_until_when_descending=False):
#
#         assert limit is None or limit >= 1, limit
#
#         if stored_entity_id not in self._by_stored_entity_id:
#             return []
#         else:
#             # Get a copy of the list of stored events for this entity.
#             stored_events = self._by_stored_entity_id[stored_entity_id][:]
#
#             # Stored event, here, are in ascending order (because they get appended to a list).
#             if not query_ascending:
#                 stored_events.reverse()
#
#             # Get timestamps (floats) from the UUID hex strings (chronologically to compare events).
#             if after is None:
#                 after_timestamp = None
#             else:
#                 after_timestamp = timestamp_from_uuid(after)
#             if until is None:
#                 until_timestamp = None
#             else:
#                 until_timestamp = timestamp_from_uuid(until)
#
#             # Start counting events (needed to stop when limit is reached).
#             count = 0
#
#             # Initialise the query results.
#             query_results = []
#
#             # Iterate over the stored events, excluding things that don't match.
#             for event in stored_events:
#                 event_timestamp = timestamp_from_uuid(event.event_id)
#
#                 if limit is not None and count >= limit:
#                     break
#
#                 # Exclude if earlier than the 'after' time.
#                 if after_timestamp:
#                     if query_ascending and not include_after_when_ascending:
#                         if event_timestamp <= after_timestamp:
#                             continue
#                     else:
#                         if event_timestamp < after_timestamp:
#                             continue
#
#                 # Exclude if later than the 'until' time.
#                 if until_timestamp:
#                     if query_ascending or include_until_when_descending:
#                         if event_timestamp > until_timestamp:
#                             continue
#                     else:
#                         if event_timestamp >= until_timestamp:
#                             continue
#
#                 query_results.append(event)
#                 count += 1
#
#                 if results_ascending != query_ascending:
#                     query_results.reverse()
#
#             return query_results



from eventsourcing.infrastructure.popo.factory import PopoInfrastructureFactory
from eventsourcing.infrastructure.popo.mapper import SequencedItemMapperForPopo
from eventsourcing.tests.sequenced_item_tests import base


class PopoTestCase(object):
    infrastructure_factory_class = PopoInfrastructureFactory
    sequenced_item_mapper_class = SequencedItemMapperForPopo

    contiguous_record_ids = True

    def construct_datastore(self):
        pass


class TestPopoRecordManagerWithIntegerSequences(PopoTestCase, base.IntegerSequencedRecordTestCase):
    def construct_record_manager(self):
        return self.construct_entity_record_manager()


class TestPopoRecordManagerWithoutContiguousRecordIDs(PopoTestCase, base.IntegerSequencedRecordTestCase):
    contiguous_record_ids = False

    def construct_record_manager(self):
        return self.construct_entity_record_manager()

#
# class TestPopoRecordManagerWithTimestampSequences(PopoTestCase, base.TimestampSequencedItemTestCase):
#     def construct_record_manager(self):
#         return self.construct_timestamp_sequenced_record_manager()
#

# class WithPopoRecordManagers(PopoTestCase, WithActiveRecordManagers):
#     def construct_entity_record_manager(self):
#         return construct_integer_sequenced_record_manager()
#
#     def construct_log_record_manager(self):
#         return construct_timestamp_sequenced_record_manager()
#
#     def construct_snapshot_record_manager(self):
#         return construct_snapshot_record_manager()
#

# class TestSimpleSequencedItemIteratorWithPopo(WithPopoRecordManagers,
#                                                 SimpleSequencedItemteratorTestCase):
#     pass
#
#
# class TestThreadedSequencedItemIteratorWithPopo(WithPopoRecordManagers,
#                                                   ThreadedSequencedItemIteratorTestCase):
#     pass



from eventsourcing.tests.datastore_tests.test_cassandra import CassandraDatastoreTestCase
from eventsourcing.tests.sequenced_item_tests import base


class TestCassandraRecordManagerWithIntegerSequences(CassandraDatastoreTestCase,
                                                     base.IntegerSequencedRecordTestCase):
    pass


class TestCassandraRecordManagerWithTimestampSequences(CassandraDatastoreTestCase,
                                                       base.TimestampSequencedItemTestCase):
    pass


class WithCassandraRecordManagers(CassandraDatastoreTestCase, base.WithRecordManagers):
    pass


class TestSequencedItemIteratorWithCassandra(WithCassandraRecordManagers,
                                             base.SequencedItemIteratorTestCase):
    pass


class TestThreadedSequencedItemIteratorWithCassandra(WithCassandraRecordManagers,
                                                     base.ThreadedSequencedItemIteratorTestCase):
    pass



from abc import ABC

from eventsourcing.application.policies import PersistencePolicy
from eventsourcing.domain.model.entity import VersionedEntity
from eventsourcing.domain.model.events import Logged
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.example.domainmodel import create_new_example
from eventsourcing.example.infrastructure import ExampleRepository
from eventsourcing.infrastructure.eventstore import EventStore
from eventsourcing.infrastructure.sequenceditemmapper import SequencedItemMapper
from eventsourcing.infrastructure.snapshotting import EventSourcedSnapshotStrategy
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder


# Please note, the code is this module is basically old-fashioned, and will
# be removed when the tests that depend on it are rewritten to use the new
# application classes in package eventsourcing.application. These were the
# original base classes for event sourced applications, but were set aside
# in favour of the new, less complicated application base classes


class ApplicationWithEventStores(ABC):
    """
    Event sourced application object class.

    Can construct event stores using given database records.
    Supports three different event stores: for log events,
    for entity events, and for snapshot events.
    """

    def __init__(self, entity_record_manager=None,
                 log_record_manager=None,
                 snapshot_record_manager=None,
                 cipher=None,
                 sequenced_item_mapper_class=SequencedItemMapper):

        self.entity_event_store = None
        if entity_record_manager:
            self.entity_event_store = self.construct_event_store(
                event_sequence_id_attr='originator_id',
                event_position_attr='originator_version',
                record_manager=entity_record_manager,
                cipher=cipher,
                sequenced_item_mapper_class=sequenced_item_mapper_class,
            )

        self.log_event_store = None
        if log_record_manager:
            self.log_event_store = self.construct_event_store(
                event_sequence_id_attr='originator_id',
                event_position_attr='timestamp',
                record_manager=log_record_manager,
                cipher=cipher,
                sequenced_item_mapper_class=sequenced_item_mapper_class,
            )

        self.snapshot_event_store = None
        if snapshot_record_manager:
            self.snapshot_event_store = self.construct_event_store(
                event_sequence_id_attr='originator_id',
                event_position_attr='originator_version',
                record_manager=snapshot_record_manager,
                cipher=cipher,
                sequenced_item_mapper_class=sequenced_item_mapper_class,
            )

    def construct_event_store(self, sequenced_item_mapper_class, event_sequence_id_attr, event_position_attr,
                              record_manager, cipher=None):
        sequenced_item_mapper = self.construct_sequenced_item_mapper(
            sequenced_item_mapper_class=sequenced_item_mapper_class,
            sequenced_item_class=record_manager.sequenced_item_class,
            event_sequence_id_attr=event_sequence_id_attr,
            event_position_attr=event_position_attr,
            cipher=cipher,
        )
        event_store = EventStore(
            record_manager=record_manager,
            sequenced_item_mapper=sequenced_item_mapper,
        )
        return event_store

    def construct_sequenced_item_mapper(self,
                                        sequenced_item_mapper_class,
                                        sequenced_item_class,
                                        event_sequence_id_attr,
                                        event_position_attr,
                                        cipher=None,
                                        json_encoder_class=ObjectJSONEncoder,
                                        json_decoder_class=ObjectJSONDecoder):
        return sequenced_item_mapper_class(
            sequenced_item_class=sequenced_item_class,
            sequence_id_attr_name=event_sequence_id_attr,
            position_attr_name=event_position_attr,
            json_encoder_class=json_encoder_class,
            json_decoder_class=json_decoder_class,
            cipher=cipher,
        )

    def close(self):
        self.entity_event_store = None
        self.log_event_store = None
        self.snapshot_event_store = None

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


class ApplicationWithPersistencePolicies(ApplicationWithEventStores):
    def __init__(self, **kwargs):
        super(ApplicationWithPersistencePolicies, self).__init__(**kwargs)
        self.entity_persistence_policy = self.construct_entity_persistence_policy()
        self.snapshot_persistence_policy = self.construct_snapshot_persistence_policy()
        self.log_persistence_policy = self.construct_log_persistence_policy()

    def construct_entity_persistence_policy(self):
        if self.entity_event_store:
            return PersistencePolicy(
                event_store=self.entity_event_store,
                persist_event_type=VersionedEntity.Event,
            )

    def construct_snapshot_persistence_policy(self):
        if self.snapshot_event_store:
            return PersistencePolicy(
                event_store=self.snapshot_event_store,
                persist_event_type=Snapshot,
            )

    def construct_log_persistence_policy(self):
        if self.log_event_store:
            return PersistencePolicy(
                event_store=self.log_event_store,
                persist_event_type=Logged,
            )

    def close(self):
        if self.entity_persistence_policy is not None:
            self.entity_persistence_policy.close()
            self.entity_persistence_policy = None
        if self.snapshot_persistence_policy is not None:
            self.snapshot_persistence_policy.close()
            self.snapshot_persistence_policy = None
        if self.log_persistence_policy is not None:
            self.log_persistence_policy.close()
            self.log_persistence_policy = None
        super(ApplicationWithPersistencePolicies, self).close()


class ExampleApplication(ApplicationWithPersistencePolicies):
    """
    Example event sourced application with entity factory and repository.
    """

    def __init__(self, **kwargs):
        super(ExampleApplication, self).__init__(**kwargs)
        self.snapshot_strategy = None
        if self.snapshot_event_store:
            self.snapshot_strategy = EventSourcedSnapshotStrategy(
                snapshot_store=self.snapshot_event_store,
            )
        assert self.entity_event_store is not None
        self.example_repository = ExampleRepository(
            event_store=self.entity_event_store,
            snapshot_strategy=self.snapshot_strategy,
        )

    def create_new_example(self, foo='', a='', b=''):
        """Entity object factory."""
        return create_new_example(foo=foo, a=a, b=b)


def construct_example_application(**kwargs):
    """Application object factory."""
    return ExampleApplication(**kwargs)


# "Global" variable for single instance of application.
_application = None


def init_example_application(**kwargs):
    """
    Constructs single global instance of application.

    To be called when initialising a worker process.
    """
    global _application
    if _application is not None:
        raise AssertionError("init_example_application() has already been called")
    _application = construct_example_application(**kwargs)


def get_example_application():
    """
    Returns single global instance of application.

    To be called when handling a worker request, if required.
    """
    if _application is None:
        raise AssertionError("init_example_application() must be called first")
    assert isinstance(_application, ExampleApplication)
    return _application


def close_example_application():
    """
    Shuts down single global instance of application.

    To be called when tearing down, perhaps between tests, in order to allow a
    subsequent call to init_example_application().
    """
    global _application
    if _application is not None:
        _application.close()
    _application = None



import os
import platform

from setuptools import find_packages, setup

from eventsourcing import __version__

is_pypy = platform.python_implementation() == 'PyPy'

# Read the docs doesn't need to build the Cassandra driver (and can't).
if 'READTHEDOCS' in os.environ:
    os.environ['CASS_DRIVER_NO_CYTHON'] = '1'

install_requires = [
    'python-dateutil<=2.8.99999',
    'pycryptodome<=3.8.99999',
    'requests<=2.22.99999',
    'readerwriterlock<=1.0.99999',
]

sqlalchemy_requires = [
    'sqlalchemy<=1.3.99999,>=0.9',
    'sqlalchemy-utils<=0.34.99999',
]

cassandra_requires = [
    'cassandra-driver<=3.18.99999'
]

django_requires = [
    'django<=2.2.99999'
]

testing_requires = cassandra_requires + sqlalchemy_requires + django_requires + [
    'mock<=3.0.99999',
    'flask<=1.1.99999',
    'flask_sqlalchemy<=2.4.99',
    'uwsgi<=2.0.99999',
    'redis<=3.2.99999',
    'celery<=4.3.99999',
    'pymysql<=0.9.99999',
    'thespian<=3.9.99999',
    # Tests use Django with PostgreSQL.
    'psycopg2cffi<=2.8.99999' if is_pypy else 'psycopg2-binary<=2.7.99999'
]

docs_requires = testing_requires + [
    'Sphinx==1.8.5',
    'python_docs_theme',
    'sphinx_py3doc_enhanced_theme',
    'sphinx_rtd_theme==0.4.3',
    'Alabaster',
    'sphinx-autobuild'
]

long_description = """
A library for event sourcing in Python.

`Package documentation is now available <http://eventsourcing.readthedocs.io/>`_.

`Please raise issues on GitHub <https://github.com/johnbywater/eventsourcing/issues>`_.
"""

packages = find_packages(
    exclude=[
        "docs",
        # "eventsourcing.contrib*",
        # "eventsourcing.tests*"
    ]
)

setup(
    name='eventsourcing',
    version=__version__,
    description='Event sourcing in Python',
    author='John Bywater',
    author_email='john.bywater@appropriatesoftware.net',
    url='https://github.com/johnbywater/eventsourcing',
    license='BSD-3-Clause',
    packages=packages,
    install_requires=install_requires,
    extras_require={
        'cassandra': cassandra_requires,
        'sqlalchemy': sqlalchemy_requires,
        'django': django_requires,
        'test': testing_requires,
        'tests': testing_requires,
        'testing': testing_requires,
        'dev': testing_requires,
        'develop': testing_requires,
        'development': testing_requires,
        'docs': docs_requires,
    },
    zip_safe=False,
    long_description=long_description,
    keywords=['event sourcing', 'event store', 'domain driven design', 'ddd', 'cqrs', 'cqs'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # 'Programming Language :: Python :: 3.5',   # we use f-strings
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from __future__ import absolute_import, division, print_function, unicode_literals

import json

import requests

from eventsourcing.application.notificationlog import Section, AbstractNotificationLog, LocalNotificationLog
from eventsourcing.utils.transcoding import ObjectJSONDecoder, ObjectJSONEncoder, json_dumps


# These classes are imported here only to avoid breaking backwards compatibility.
# Todo: Remove this import statement >= v8.x
from eventsourcing.application.notificationlog import RecordManagerNotificationLog, BigArrayNotificationLog


class RemoteNotificationLog(AbstractNotificationLog):
    """
    Presents notification log sections retrieved an HTTP API
    that presents notification log sections in JSON format,
    for example by using a NotificationLogView.
    """
    def __init__(self, base_url, json_decoder_class=None):
        """
        Initialises remote notification log object.

        :param str base_url: A URL for the HTTP API.
        :param JSONDecoder json_decoder_class: JSON decoder class used to decode remote sections.
        """
        self.base_url = base_url
        self.json_decoder_class = json_decoder_class

    def __getitem__(self, section_id):
        """
        Returns a section of notification log.

        :param str section_id: ID of the section of the notification log.
        :return: Identified section of notification log.
        :rtype: Section
        """
        section_json = self.get_json(section_id)
        return self.deserialize_section(section_json)

    def deserialize_section(self, section_json):
        try:
            decoder_class = self.json_decoder_class or ObjectJSONDecoder
            section = Section(**json.loads(section_json, cls=decoder_class))
        except ValueError as e:
            raise ValueError("Couldn't deserialize notification log section: "
                             "{}: {}".format(e, section_json))
        return section

    def get_json(self, section_id):
        notification_log_url = self.make_notification_log_url(section_id)
        return self.get_resource(notification_log_url)

    def get_resource(self, doc_url):
        representation = requests.get(doc_url).content
        if isinstance(representation, type(b'')):
            representation = representation.decode('utf8')
        return representation

    def make_notification_log_url(self, section_id):
        return '{}/{}/'.format(self.base_url.strip('/'), section_id)


class NotificationLogView(object):
    """
    Presents sections of a notification log in JSON format.

    Can be used to make an HTTP API that can be used
    remotely, for example by a RemoteNotificationLog.
    """
    def __init__(self, notification_log: LocalNotificationLog, json_encoder_class=None):
        """
        Initialises notification log view object.

        :param LocalNotificationLog notification_log: A notification log object
        :param JSONEncoder json_encoder_class: JSON encoder class
        """
        assert isinstance(notification_log, LocalNotificationLog), type(notification_log)
        self.notification_log = notification_log
        self.json_encoder_class = json_encoder_class or ObjectJSONEncoder

    def present_section(self, section_id):
        """

        Returns a section of notification log in JSON format.

        :param section_id: ID of the section of the notification log.
        :return: Identified section of notification log in JSON format.

        :rtype: str
        """
        section = self.notification_log[section_id]
        is_archived = bool(section.next_id)
        section_json = json_dumps(section.__dict__, self.json_encoder_class)
        return section_json, is_archived



from abc import abstractmethod
from threading import Lock


class AbstractIntegerSequenceGenerator(object):

    def __iter__(self):
        return self

    @abstractmethod
    def __next__(self):
        """
        Returns the next item in the container.
        """


class SimpleIntegerSequenceGenerator(AbstractIntegerSequenceGenerator):

    def __init__(self, i=0):
        self.i = i
        self.lock = Lock()

    def __next__(self):
        self.lock.acquire()
        i = self.i
        self.i += 1
        self.lock.release()
        return i



from unittest.case import TestCase
from uuid import uuid4

from cassandra import InvalidRequest
from cassandra.cluster import NoHostAvailable
from cassandra.cqlengine import CQLEngineException

from eventsourcing.exceptions import DatasourceSettingsError
from eventsourcing.infrastructure.cassandra.factory import CassandraInfrastructureFactory
from eventsourcing.infrastructure.cassandra.records import IntegerSequencedRecord, SnapshotRecord, \
    TimestampSequencedRecord
from eventsourcing.infrastructure.cassandra.datastore import CassandraDatastore, CassandraSettings
from eventsourcing.infrastructure.datastore import DatastoreConnectionError, DatastoreTableError
from eventsourcing.tests.datastore_tests import base

DEFAULT_KEYSPACE_FOR_TESTING = 'eventsourcing_tests'


class CassandraDatastoreTestCase(base.AbstractDatastoreTestCase):
    """
    Uses the datastore object to set up connection to and tables in Cassandra.
    """
    infrastructure_factory_class = CassandraInfrastructureFactory

    def construct_datastore(self):
        return CassandraDatastore(
            settings=CassandraSettings(default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING),
            tables=(IntegerSequencedRecord, TimestampSequencedRecord, SnapshotRecord),
        )


class TestCassandraDatastore(CassandraDatastoreTestCase, base.DatastoreTestCase):
    def list_records(self):
        try:
            return list(IntegerSequencedRecord.objects.all())
        except (CQLEngineException, NoHostAvailable) as e:
            raise DatastoreConnectionError(e)
        except InvalidRequest as e:
            raise DatastoreTableError(e)

    def create_record(self):
        record = IntegerSequencedRecord(
            sequence_id=uuid4(),
            position=0,
            topic='topic',
            state='{}'
        )
        try:
            record.save()
        except (CQLEngineException, NoHostAvailable) as e:
            raise DatastoreConnectionError(e)
        except InvalidRequest as e:
            raise DatastoreTableError(e)
        return record


class TestPlainTextAuthProvider(TestCase):
    def test_plain_text_auth_provider(self):
        datastore = CassandraDatastore(
            settings=CassandraSettings(
                default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING,
                username='username',
                password='password',
            ),
            tables=(IntegerSequencedRecord,),
        )
        datastore.setup_connection()


class TestDatabaseSettingsError(TestCase):
    def test_consistency_level_error(self):
        datastore = CassandraDatastore(
            settings=CassandraSettings(
                default_keyspace=DEFAULT_KEYSPACE_FOR_TESTING,
                consistency='invalid',
            ),
            tables=(IntegerSequencedRecord,),
        )

        with self.assertRaises(DatasourceSettingsError):
            datastore.setup_connection()



from unittest import skip

from eventsourcing.application.popo import PopoApplication

from eventsourcing.tests.sequenced_item_tests.test_popo_record_manager import PopoTestCase
from eventsourcing.tests.test_process import TestProcessApplication


class TestProcessWithPopos(PopoTestCase, TestProcessApplication):
    process_class = PopoApplication

    @skip("Popo record manager doesn't support pipelines")
    def test_causal_dependencies(self):
        super(TestProcessWithPopos, self).test_causal_dependencies()


del TestProcessApplication



import logging
import os

from eventsourcing.application.process import ProcessApplication
from eventsourcing.domain.model.aggregate import BaseAggregateRoot
from eventsourcing.domain.model.decorators import retry
from eventsourcing.exceptions import OperationalError, RecordConflictError
from eventsourcing.tests.test_process import ExampleAggregate


def set_db_uri():
    host = os.getenv('MYSQL_HOST', '127.0.0.1')
    user = os.getenv('MYSQL_USER', 'root')
    password = os.getenv('MYSQL_PASSWORD', '')
    db_uri = 'mysql+pymysql://{}:{}@{}/eventsourcing?charset=utf8mb4&binary_prefix=true'.format(user, password, host)
    os.environ['DB_URI'] = db_uri


class Order(BaseAggregateRoot):
    def __init__(self, **kwargs):
        super(Order, self).__init__(**kwargs)
        self.is_reserved = False
        self.is_paid = False

    class Event(BaseAggregateRoot.Event): pass

    class Created(Event, BaseAggregateRoot.Created): pass

    class Reserved(Event):
        def mutate(self, order):
            assert not order.is_reserved, "Order {} already reserved.".format(order.id)
            order.is_reserved = True
            order.reservation_id = self.reservation_id

    class Paid(Event):
        def mutate(self, order):
            assert not order.is_paid, "Order {} already paid.".format(order.id)
            order.is_paid = True
            order.payment_id = self.payment_id

    def set_is_reserved(self, reservation_id):
        self.__trigger_event__(Order.Reserved, reservation_id=reservation_id)

    def set_is_paid(self, payment_id):
        self.__trigger_event__(self.Paid, payment_id=payment_id)


class Reservation(BaseAggregateRoot):
    def __init__(self, order_id, **kwargs):
        super(Reservation, self).__init__(**kwargs)
        self.order_id = order_id

    class Event(BaseAggregateRoot.Event): pass

    class Created(Event, BaseAggregateRoot.Created): pass

    @classmethod
    def create(cls, order_id):
        return cls.__create__(order_id=order_id)


class Payment(BaseAggregateRoot):
    def __init__(self, order_id, **kwargs):
        super(Payment, self).__init__(**kwargs)
        self.order_id = order_id

    class Event(BaseAggregateRoot.Event): pass

    class Created(Event, BaseAggregateRoot.Created): pass

    @classmethod
    def make(self, order_id):
        return self.__create__(order_id=order_id)


@retry((OperationalError, RecordConflictError), max_attempts=10, wait=0.01)
def create_new_order():
    order = Order.__create__()
    order.__save__()
    return order.id


logger = logging.getLogger()


class Orders(ProcessApplication):
    persist_event_type = Order.Created

    @staticmethod
    def policy(repository, event):
        if isinstance(event, Reservation.Created):
            # Set the order as reserved.
            order = repository[event.order_id]
            assert not order.is_reserved
            order.set_is_reserved(event.originator_id)
            # logger.info('set Order as reserved')
            # print(f'set Order {event.order_id} as reserved')

        elif isinstance(event, Payment.Created):
            # Set the order as paid.
            order = repository[event.order_id]
            assert not order.is_paid
            order.set_is_paid(event.originator_id)
            # logger.info('set Order as paid')
            # print(f'set Order {event.order_id} as paid')


class Reservations(ProcessApplication):
    @staticmethod
    def policy(repository, event):
        if isinstance(event, Order.Created):
            # Create a reservation.
            # time.sleep(0.5)
            # logger.info('created Reservation for order')
            # print(f'created Reservation for order {event.originator_id}')
            return Reservation.create(order_id=event.originator_id)


class Payments(ProcessApplication):
    @staticmethod
    def policy(repository, event):
        if isinstance(event, Order.Reserved):
            # Make a payment.
            # time.sleep(0.5)
            # logger.info('created Payment for order')
            # print(f'created Payment for order {event.originator_id}')
            return Payment.make(order_id=event.originator_id)


class Examples(ProcessApplication):
    persist_event_type = ExampleAggregate.Created

    @staticmethod
    def policy(repository, event):
        if isinstance(event, ExampleAggregate.Created):
            repository[event.originator_id].move_on()



from abc import ABC, abstractmethod
from threading import Thread

from eventsourcing.infrastructure.base import AbstractSequencedItemRecordManager


class AbstractSequencedItemIterator(ABC):
    DEFAULT_PAGE_SIZE = 1000

    def __init__(self, record_manager, sequence_id, page_size=None, gt=None, gte=None, lt=None, lte=None,
                 limit=None, is_ascending=True):
        assert isinstance(record_manager, AbstractSequencedItemRecordManager), type(record_manager)
        assert isinstance(page_size, (int, type(None)))
        assert isinstance(limit, (int, type(None)))
        self.record_manager = record_manager
        self.sequence_id = sequence_id
        self.page_size = page_size or self.DEFAULT_PAGE_SIZE
        self.gt = gt
        self.gte = gte
        self.lte = lte
        self.lt = lt
        self.limit = limit
        self.query_counter = 0
        self.page_counter = 0
        self.all_item_counter = 0
        self.is_ascending = is_ascending
        self._position = None

    def _inc_page_counter(self):
        """
        Increments the page counter.

        Each query result as a page, even if there are no items in the page. This really counts queries.
         - it is easy to divide the number of events by the page size if the "correct" answer is required
         - there will be a difference in the counts when the number of events can be exactly divided by the page
           size, because there is no way to know in advance that a full page is also the last page.
        """
        self.page_counter += 1

    def _inc_query_counter(self):
        """
        Increments the query counter.
        """
        self.query_counter += 1

    def _inc_all_event_counter(self):
        self.all_item_counter += 1

    def _update_position(self, sequenced_item):
        assert isinstance(sequenced_item, self.record_manager.sequenced_item_class), type(sequenced_item)
        self._position = getattr(sequenced_item, self.record_manager.field_names.position)

    @abstractmethod
    def __iter__(self):
        """
        Yields a continuous sequence of items.
        """


class SequencedItemIterator(AbstractSequencedItemIterator):
    def __iter__(self):
        """
        Yields a continuous sequence of items from "pages"
        of sequenced items retrieved using the record manager.
        """
        gt = self.gt
        gte = self.gte
        lt = self.lt
        lte = self.lte

        while True:
            # Get next page of events.
            if self.limit is not None:
                limit = min(self.page_size, self.limit - self.all_item_counter)
            else:
                limit = self.page_size

            if limit == 0:
                return

            # Get the events.
            if self._position is not None:
                if self.is_ascending:
                    gt = self._position
                    gte = None
                else:
                    lt = self._position
                    lte = None

            sequenced_items = self.record_manager.get_items(
                sequence_id=self.sequence_id,
                gt=gt,
                gte=gte,
                lt=lt,
                lte=lte,
                limit=limit,
                query_ascending=self.is_ascending,
                results_ascending=self.is_ascending,
            )

            self._inc_query_counter()

            # Start counting events in this page.
            page_item_counter = 0

            # Yield each stored event.
            for sequenced_item in sequenced_items:

                # Count each event.
                self._inc_all_event_counter()
                page_item_counter += 1

                # Yield the event.
                yield sequenced_item

                # Remember the position as the last event.
                self._update_position(sequenced_item)

            # If that wasn't an empty page, count the page.
            if page_item_counter:
                self._inc_page_counter()

            # If that wasn't a full page, stop iterating (there can be no more items).
            if page_item_counter != self.page_size:
                return


class ThreadedSequencedItemIterator(AbstractSequencedItemIterator):
    def __iter__(self):
        # Start a thread to get a page of events.
        thread = self.start_thread()

        # Get pages of stored events, until the page isn't full.
        while True:
            # Wait for the next page of events.
            thread.join(timeout=30)

            # Count the query.
            self._inc_query_counter()

            # Get the stored events from the thread.
            stored_events = thread.stored_events

            # Count the number of stored events that were retrieved.
            num_stored_events = len(stored_events)

            if num_stored_events:
                self._inc_page_counter()

            # Decide if this is the last page.
            is_last_page = (
                               num_stored_events != self.page_size
                           ) or (
                               self.all_item_counter + num_stored_events == self.limit
                           )

            if not is_last_page:
                # Update loop variables.
                position = stored_events[-1]
                self._update_position(position)

                # Start the next thread.
                thread = self.start_thread()

            # Yield each stored event.
            for stored_event in stored_events:

                # Stop if we're over the limit.
                if self.limit and self.all_item_counter >= self.limit:
                    return

                # Count each event.
                self._inc_all_event_counter()

                # Yield the event.
                yield stored_event

            # If that was the last page, then stop iterating.
            if is_last_page:
                return

    def start_thread(self):
        gt = self.gt
        gte = self.gte
        lt = self.lt
        lte = self.lte

        if self._position is not None:
            if self.is_ascending:
                gt = self._position
                gte = None
            else:
                lt = self._position
                lte = None

        thread = GetEntityEventsThread(
            record_manager=self.record_manager,
            sequence_id=self.sequence_id,
            gt=gt,
            gte=gte,
            lt=lt,
            lte=lte,
            page_size=self.page_size,
            is_ascending=self.is_ascending
        )
        thread.start()
        return thread


class GetEntityEventsThread(Thread):
    def __init__(self, record_manager, sequence_id, gt=None, gte=None, lt=None, lte=None, page_size=None,
                 is_ascending=True, *args, **kwargs):
        super(GetEntityEventsThread, self).__init__(*args, **kwargs)
        assert isinstance(record_manager, AbstractSequencedItemRecordManager), type(record_manager)
        self.record_manager = record_manager
        self.stored_entity_id = sequence_id
        self.gt = gt
        self.gte = gte
        self.lt = lt
        self.lte = lte
        self.page_size = page_size
        self.is_ascending = is_ascending
        self.stored_events = None

    def run(self):
        self.stored_events = list(self.record_manager.get_items(
            sequence_id=self.stored_entity_id,
            gt=self.gt,
            gte=self.gte,
            lt=self.lt,
            lte=self.lte,
            limit=self.page_size,
            query_ascending=self.is_ascending,
            results_ascending=self.is_ascending,
        ))
